/* fparser.c
**
** Parser for an e-mail filter
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "prefs.h"
#include "tools.h"
#include "inline-funcs.h"
#include "mrmacs.h"
#include "rfc822_address.h"

#ifndef DEFHOST
# include "userdb.h"
# define DEFHOST mailhost()
#endif

#include "fparser.h"

static int lc = 0;

/* Hole die nchste Zeile aus der Filterdatei. Wenn `noskip' nicht gesetzt ist,
** dann werden Leerzeilen und Zeilen, die mit `#' beginnen, bersprungen.
** Auerdem wird ein Zeilenzhler mitgefhrt, der in den folgenden Routinen
** zur Anzeige von Fehlern verwendet wird.
*/
static char *getline (FILE *ff, int noskip)
{
    static char line[1024];
    char *p;
RETRY:
    ifnull (fgets (line, 1024, ff)) { return NULL; }
    ++lc;
    if (cuteol(line) == 0) {
	for (;;) {
	    fgets (line, 1024, ff);
	    if (cuteol(line) != 0) { break; }
	}
	fprintf (stderr, "%s: F(%d) - line too long!\n", progname, lc);
	vxlog (0, "F(%d): line too long - skipping...\n", lc);
	goto RETRY;
    }
    p = line;
#if 0
    if (!noskip) { while (isws (*p)) { ++p; } }
#endif
    if (!noskip && (!*p || *p == '#')) { goto RETRY; }
    return p;
}

static void condfree (condlist_t ol);
static action_t p_actions (FILE *ff, char *vars[], const char *sender);
static condlist_t p_conditions (FILE *ff);

/* fparser ist die von diesem Modul exportierte Routine. Aus der geffneten
** Datei (Argument) wird der Filter eingelesen und parsiert; das Ergebnis
** der Funktion ist eine Struktur, die den eingelesenen Filter reprsentiert.
*/
filter_t fparser (FILE *fp, char *vars[], const char *sender)
{
    filter_t first = (filter_t) NULL, current = (filter_t) NULL;
    condlist_t cond = (condlist_t) NULL;
    action_t action = (action_t) NULL;
    char *p;
    int head = 0, body = 0, pre = 0;
    while (noNULL(p = getline (fp, 0))) {
	if (*p == ':' || (!lccmp (p, "FILTER", 6) && isws(p[6]))) {
	    if (*p == ':') { ++p; } else { p += 6; }
	    while (*p) {
		while (isws(*p)) { ++p; }
		if (*p == '#') { break; }
		if (*p == '+') { ++p; pre = 1; continue; }
		if (*p == 'H' || *p == 'h') { ++p; head = 1; continue; }
		if (*p == 'B' || *p == 'b') { ++p; body = 1; continue; }
		fprintf (stderr, "%s: F(%d) - invalid filter option `%c'.\n",
				 progname, lc, *p);
		vxlog (0, "F(%d): invalid filter option `%c'.\n", lc, *p);
		++p;
	    }
	    ifnull (cond = p_conditions (fp)) { continue; }
	    ifnull (action = p_actions (fp, vars, sender)) {
		condfree (cond); continue;
	    }
	    ifnull (first) {
		first = current = galloc (__FILE__, __LINE__,
					  sizeof(struct filter_s));
	    } else {
		current->next = galloc (__FILE__, __LINE__,
					sizeof(struct filter_s));
		current = current->next;
	    }
	    current->next = (filter_t) NULL;
	    current->head = head; current->body = body;
	    current->result = 0;
	    current->pre = pre;
	    current->conditions = cond;
	    current->action = action;
	} else {
	    fprintf (stderr, "%s: F(%d) - syntax error.\n", progname, lc);
	    vxlog (0, "F(%d): syntax error.\n", lc);
	}
    }
    return first;
}

/* Lies das Steuerzeichen fr die Unterscheidung zwischen Gro- und
** Kleinschreibung und gib je nachdem, ob es `i' oder ` ' war,
** REG_ICASE oder 0 zurck.
*/
static int rxcase (char **_p)
{
    char *p = *_p;
    int icase = 0;
    if (*p == 'i' || *p == 'I') {
	icase = REG_ICASE;
    } else if (isws (*p)) {
	icase = 0;
    } else {
	fprintf (stderr, "%s: F(%d) - need either ` ' or `i' before regex.\n",
		 progname, lc);
	vxlog (0, "F(%d): need either ` ' or `i' beford regex.\n", lc);
	return -1;
    }
    ++p; while (isws (*p)) { ++p; }
    *_p = p;
    return icase;
}

/* Hilfsroutine zur Freigabe des Speicherplatzes fr die `UND'-verknpften
** Filterbedingungen.
*/
static void alsofree (condlist_t al)
{
    condlist_t nx;
    while (noNULL(al)) {
	nx = al->also;
	regfree (&al->rx);
	gfree (al);
	al = nx;
    }
}

/* Hilfroutine zur Freigabe des Speicherplatzes fr die `ODER'-verknpften
** Filterbedingungen.
*/
static void condfree (condlist_t ol)
{
    condlist_t nx;
    while (noNULL(ol)) {
	nx = ol->next;
	alsofree (ol->also);
	regfree (&ol->rx);
	gfree (ol);
	ol = nx;
    }
}

#define eol(c) ((c) == '\r' || (c) == '\n' || isws((c)) || (c) == '#')

/* Parsiere den Bedingungsteil eines Filters.
*/
static condlist_t p_conditions (FILE *ff)
{
    condlist_t o1st = (condlist_t) NULL, ocurrent = (condlist_t) NULL,
	       acurrent = (condlist_t) NULL;
    char *p;
    int icase = 0, rxerr, andxp = 0;
    regex_t rx;
    while (noNULL(p = getline (ff, 0))) {
	if (!strncmp (p, "==>", 3)) { break; }
	if (*p == ';' || !lccmp(p, "END", -1)) { p = NULL; break; }
	andxp = 0;
	if (*p == '+') { ++p; andxp = 1; }
	if ((icase = rxcase (&p)) < 0) { continue; }
	if (!*p) {
	    fprintf (stderr, "%s: F(%d) - empty regex.\n", progname, lc);
	    vxlog (0, "F(%d): empty regex.\n", lc);
	    continue;
	}
	if ((rxerr = regcomp (&rx, p, REG_NOSUB|REG_EXTENDED|icase))) {
	    char rxemsg[256];
	    regerror (rxerr, &rx, rxemsg, sizeof(rxemsg));
	    fprintf (stderr, "%s: F(%d) - %s\n", progname, lc, rxemsg);
	    vxlog (0, "F(%d): %s\n", lc, rxemsg);
	    continue;
	}
	if (andxp && noNULL(ocurrent)) {
	    ifnull (acurrent) {
		ocurrent->also = acurrent = galloc (__FILE__, __LINE__,
						    sizeof(struct condlist_s));
	    } else {
		acurrent->also = galloc (__FILE__, __LINE__,
					 sizeof(struct condlist_s));
		acurrent = acurrent->also;
	    }
	    acurrent->next = acurrent->also = (condlist_t) NULL;
	    acurrent->alsores = 0;
	    acurrent->rx = rx;
	    continue;
	}
	ifnull (o1st) {
	    o1st = ocurrent = (condlist_t) galloc (__FILE__, __LINE__,
						   sizeof(struct condlist_s));
	} else {
	    ocurrent->next = (condlist_t) galloc (__FILE__, __LINE__,
						  sizeof(struct condlist_s));
	    ocurrent = ocurrent->next;
	}
	ocurrent->next = ocurrent->also = (condlist_t) NULL;
	ocurrent->alsores = 0;
	ocurrent->rx = rx;
    }
    ifnull (p) {
	fprintf (stderr, "%s: F(%d) - missing action part\n", progname, lc);
	vxlog (0, "F(%d): missing action part\n", lc);
	condfree (o1st);
	return (condlist_t) NULL;
    }
    ifnull (o1st) {
	fprintf (stderr, "%s: F(%d) - missing filter conditions\n",
			 progname, lc);
	vxlog (0, "F(%d); missing filter conditions\n", lc);
	return (condlist_t) NULL;
    }
    return o1st;
}

/* Namen => Aktionen-Zuordnungen */
static struct acs {
    char *name;
    acmode_t action;
} aclist[] = {
    { "delete", ac_delete },
    { "drop", ac_drop },
    { "exec", ac_exec },
    { "forwardto", ac_forward },
    { "insert", ac_insert },
    { "mailbox", ac_mailbox },
    { "mailto", ac_mailto },
    { "replace", ac_replace },
    { NULL, ac_none }
};

static int p_headel (char **_p, char **_elem, int *_ix);
static char *p_addr (char **_p);
static char *p_mbox (char **_p);

/* Substituiere alle Variablen, deren `Name=Wert'-Paare im zweiten Argument
** bergeben wurden in der als erstes Argument bergebenen Zeichenkette.
** Gib das entsprechend modifizierte erste Argument zurck.
** =================================================================
** Das erste Argument mu eine dynamisch erzeugte Zeichenkette sein.
** =================================================================
*/
static char *insvars (char *in, char *vars[])
{
    char *name, *value, *keep, *res = in;
    int ix;
    size_t len = 0, l;
    ifnull (in) { return NULL; }
    for (ix = 0; noNULL(vars[ix]); ++ix) {
	l = strlen (vars[ix]);
	if (l > len) { len = l; }
    }
    if (len == 0) { return NULL; }
    name = galloc (__FILE__, __LINE__, len + 1);
    for (ix = 0; noNULL(vars[ix]); ++ix) {
	strcpy(name, vars[ix]);
	ifnull (value = strchr (name, '=')) {
	    keep = gvarins (__FILE__, __LINE__, res, name, "");
	} else {
	    *value++ = '\0';
	    keep = gvarins (__FILE__, __LINE__, res, name, value);
	}
	gfree (res);
	res = keep;
    }
    gfree (name);
    return res;
}

static char *findval (char **vars, char *name)
{
    int ix;
    size_t len = strlen (name);
    for (ix = 0; noNULL(vars[ix]); ++ix) {
	if (!strncmp (vars[ix], name, len) && vars[ix][len] == '=') {
	    return &vars[ix][len + 1];
	}
    }
    return NULL;
}
	    

/* Parsiere den Aktionsteil eines Filters */
static action_t p_actions (FILE *ff, char *vars[], const char *sender)
{
    action_t a1st = (action_t) NULL, acurrent = (action_t) NULL;
    acmode_t acm = ac_none;
    struct acs *acl;
    int len = 0, ix, noread;
    struct action_s action;
    char *msg = NULL, *msg1, *p, *q, *r;
    while (noNULL(p = getline (ff, 0))) {
p_actions_again:
	noread = 0;
	while (isws (*p)) { ++p; }
	if (*p == ';' || !lccmp (p, "END", -1)) { break; }
	acm = ac_none;
	for (ix = 0; noNULL((acl = &aclist[ix])->name); ++ix) {
	    len = strlen (acl->name);
	    if (!lccmp (acl->name, p, len)) { acm = acl->action; break; }
	}
	if (acm == ac_none) {
	    fprintf (stderr, "%s: F(%d) - invalid action (%s).\n",
			     progname, lc, p);
	    continue;
	}
	p += len;
	if (*p && !isws (*p)) {
	    fprintf (stderr, "%s: F(%d) - invalid action.\n",
			     progname, lc);
	    vxlog (0, "F(%d): invalid action.\n", lc);
	    continue;
	}
	while (isws (*p)) { ++p; }
	action.next = (action_t) NULL;
	action.action = acm;
	switch (acm) {
	    case ac_none: break;
	    case ac_insert:
		q = p = insvars (gdupl (__FILE__, __LINE__, p), vars);
		if (p_headel (&p, &action.a.insert.elem, (int *) NULL)) {
		    gfree (q); continue;
		}
		action.a.insert.new = gdupl (__FILE__, __LINE__, p);
		gfree (q);
		break;
	    case ac_replace:
		q = p = insvars (gdupl (__FILE__, __LINE__, p), vars);
		if (p_headel(&p, &action.a.replace.elem,
				 &action.a.replace.ix)) {
		    gfree (q); continue;
		}
		action.a.insert.new = gdupl (__FILE__, __LINE__, p);
		gfree (q);
		break;
	    case ac_delete:
		q = p = insvars (gdupl (__FILE__, __LINE__, p), vars);
		if (p_headel (&p, &action.a.delete.elem, &action.a.delete.ix)) {
		    gfree (q); continue;
		}
		gfree (q);
		break;
	    case ac_mailto:
		/* Ersetze zunchst in einer dynamisch erzeugten Kopie der
		** Zeile alle vorkommenden `$(variable)' durch die (in `vars')
		** bergebenen Variablenwerte. Da `insvars' eine dynamisch
		** erzeugte Zeichenkette als erstes Argument verlangt, erzeuge
		** ich eine solche aus dem Rest der Zeile (ab `p') mittels
		** `gdupl'.
		*/
		q = p = insvars (gdupl (__FILE__, __LINE__, p), vars);
		ifnull (r = findval (vars, "TARGET")) {
		    fprintf (stderr, "%s: F(%d) - invalid TARGET value - "
				     "MailTo action ignored.\n",
				     progname, lc);
		    vxlog (0, "F(%d): invalid TARGET value - MailTo action "
			      "ignored.\n", lc);
		    gfree (q); continue;
		}
		action.a.mailto.from = gdupl (__FILE__, __LINE__, r);
		ifnull (action.a.mailto.to = p_addr (&p)) {
		    gfree (q); gfree (action.a.mailto.from); continue;
		}
		gfree (q);
		while (noNULL(p = getline (ff, 0))) {
		    if (*p != '+') { break; }
		    ++p;
		    /* Ich fge die Variablensubstitution hier ebenfalls ein,
		    ** da der Benutzer sich seine Header entsprechend setzen
		    ** knnen soll. Da `insvars' eine dynamisch erzeugte
		    ** Zeichenkette verlangt, verwende ich als erstes Argument
		    ** hier wiederum eine Kopie der Zeile ab der Position, auf
		    ** die `p' zeigt.
		    */
		    q = insvars (gdupl (__FILE__, __LINE__, p), vars);
		    ifnull (msg) {
			msg = gconc (__FILE__, __LINE__, q, "\r\n", NULL);
		    } else {
			msg1 = gconc (__FILE__, __LINE__, msg, q, "\r\n", NULL);
			gfree (msg); msg = msg1;
		    }
		    gfree (q);
		}
		action.a.mailto.msg = msg;
		noread = 1;
		break;
	    case ac_forward:
		/* Auch bei einer `ForwardTo'-Aktion fhre ich Variablen-
		** substitutionen durch - wieder auf der dynamisch erzeugten
		** Kopie der Zeile (ab `p') durch.
		*/
		q = p = insvars (gdupl (__FILE__, __LINE__, p), vars);
		action.a.forward.from = gdupl (__FILE__, __LINE__, sender);
		ifnull (action.a.forward.to = p_addr (&p)) {
		    gfree (q);continue;
		}
		gfree (q);
		break;
	    case ac_drop:
		action.a.drop.from = gdupl(__FILE__, __LINE__, sender);
		break;
	    case ac_exec:
		action.a.exec.from = gdupl(__FILE__, __LINE__, sender);
		action.a.exec.cmd = insvars(gdupl(__FILE__, __LINE__, p),
					    vars);
		break;
	    case ac_mailbox:
		ifnull (action.a.mailbox.name = p_mbox (&p)) { continue; }
		action.a.mailbox.from = gdupl(__FILE__, __LINE__, sender);
		break;
	}
	ifnull (a1st) {
	    a1st = acurrent = galloc (__FILE__, __LINE__,
				      sizeof(struct action_s));
	} else {
	    acurrent->next = galloc (__FILE__, __LINE__,
				     sizeof(struct action_s));
	    acurrent = acurrent->next;
	}
	*acurrent = action;
	if (noread) { goto p_actions_again; }
    }
    ifnull (a1st) {
	fprintf (stderr, "%s: F(%d) - missing filter action(s)\n",
			 progname, lc);
	vxlog (0, "F(%d): missing filter action(s)\n", lc);
	return (action_t) NULL;
    }
    return a1st;
}

/* Parsiere den Namen eines Header-Elementes einer E-Mail. */
static int p_headel (char **_p, char **_elem, int *_ix)
{
    char *p = *_p, *q = p, *r, *n;
    char *res; int ix = 0; long nv = -1l;
    /* scan the header element's name */
    while (*p > 31 && *p < 127 && *p != ':' && !isws (*p)) { ++p; }
    r = p;
    if (*p == ':') {
	/* scan the index of a multiple occurrency header element
	** - but only, if an index is allowed.
	*/
	++p;
	unlessnull (_ix) { nv = strtol (p, &n, 0); p = n; }
	if (*p && !isws (*p)) {
	    fprintf (stderr, "%s: F(%d) - field index must be a number or "
			     "empty.\n", progname, lc);
	    vxlog (0, "F(%d): field index is no number.\n", lc);
	    return 1;
	}
	if (nv < -1l || nv > 255l) {
	    fprintf (stderr, "%s: F(%d) - field index out of range -1..255.\n",
			     progname, lc);
	    vxlog (0, "F(%d): field index (%ld) out of range.\n", lc, nv);
	    return 1;
	}
	++p;
	ix = (int) nv;
    }
    /* skip the following white spaces */
    while (isws(*p)) { ++p; }
    *_p = p;
    res = galloc (__FILE__, __LINE__, (r - q) + 1);
    memcpy (res, q, r - q);
    res[r - q] = '\0';
    *_elem = res; unlessnull (_ix) { *_ix = ix; }
    return 0;
}

/* parse an e-mail address; use the address parser for that reason... */
static char *p_addr (char **_p)
{
    char *res;
    ifnull (res = rfc822_addrstrg (_p, DEFHOST)) {
	fprintf (stderr, "%s: F(%d) - invalid e-mail address.\n",
		 progname, lc);
	vxlog (0, "F(%d): invalid address `%s'.\n", lc, *_p);
    }
    return res;
}

/* parse the name of a mailbox. */
static char *p_mbox (char **_p)
{
    char *p = *_p, *q = p, *r, *res = NULL;
    if (!isalpha (*p)) {
	fprintf (stderr, "%s: F(%d) - invalid char `%c' in mailbox name.\n",
			 progname, lc, *p);
	vxlog (0, "F(%d): invalid char `%c' in mailbox name.\n", lc, *p);
	*_p = p;
	return NULL;
    }
    ++p;
    while (isalpha(*p) || isdigit(*p) || *p == '_' || *p == '-') { ++p; }
    if (*p && !isws(*p)) {
	fprintf (stderr, "%s: F(%d) - invalid char `%c' in mailbox name.\n",
			 progname, lc, *p);
	vxlog (0, "F(%d): invalid char `%c' in mailbox name.\n", lc, *p);
	*_p = p;
	return NULL;
    }
    r = p;
    while (isws (*p)) { ++p; }
    *_p = p;
    res = galloc (__FILE__, __LINE__, (r - q) + 1);
    memcpy (res, q, (r - q));
    res[r - q] = '\0';
    return res;
}

static void actionfree (action_t action)
{
    action_t ac = action;
    while (noNULL(ac)) {
	action = ac->next;
	switch (ac->action) {
	    case ac_none:
		break;
	    case ac_insert:
		gfree(ac->a.insert.elem);
		gfree(ac->a.insert.new);
		break;
	    case ac_replace:
		gfree(ac->a.replace.elem);
		gfree(ac->a.replace.new);
		break;
	    case ac_delete:
		gfree(ac->a.delete.elem);
		break;
	    case ac_mailto:
		gfree(ac->a.mailto.from);
		gfree(ac->a.mailto.to);
		gfree(ac->a.mailto.msg);
		break;
	    case ac_forward:
		gfree(ac->a.forward.from);
		gfree(ac->a.forward.to);
		break;
	    case ac_drop:
		gfree(ac->a.drop.from);
		break;
	    case ac_exec: break;
		gfree(ac->a.exec.from);
		gfree(ac->a.exec.cmd);
		break;
	    case ac_mailbox:
		gfree(ac->a.mailbox.from);
		gfree(ac->a.mailbox.name);
		break;
	}
	gfree (ac);
	ac = action;
    }
}

void filter_free (filter_t filter)
{
    filter_t ft = filter;
    while (noNULL(ft)) {
	filter = ft->next;
	condfree (ft->conditions);
	actionfree (ft->action);
	gfree (ft);
	ft = filter;
    }
}

const char *actype (action_t ac)
{
    switch (ac->action) {
	case ac_none:    return "None";
	case ac_insert:  return "Insert";
	case ac_replace: return "Replace";
	case ac_delete:  return "Delete";
	case ac_mailto:  return "MailTo";
	case ac_forward: return "ForwardTo";
	case ac_drop:    return "Drop";
	case ac_exec:    return "Exec";
	case ac_mailbox: return "Mailbox";
    }
    return NULL;
}
